<?php

namespace Drupal\commerce_sezzle_pay\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_price\Price;
use Drupal\commerce_price\RounderInterface;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Url;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Form\FormStateInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Drupal\commerce_product\Entity\ProductVariation;
use Drupal\commerce_order\Entity\OrderInterface;


/**
 * Provides the SezzlePay offsite Checkout payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "sezzle_pay_redirect_checkout",
 *   label = @Translation("SezzlePay (Redirect to sezzlepay)"),
 *   display_label = @Translation("SezzlePay"),
 *    forms = {
 *     "offsite-payment" =
 *   "Drupal\commerce_sezzle_pay\PluginForm\SezzlePayCheckoutForm",
 *   },
 *   payment_method_types = {"credit_card"},
 *   credit_card_types = {
 *     "amex", "discover", "mastercard", "visa",
 *   },
 * )
 */
class SezzlePayCheckout extends OffsitePaymentGatewayBase implements SezzlePayCheckoutInterface {


  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The price rounder.
   *
   * @var \Drupal\commerce_price\RounderInterface
   */
  protected $rounder;

  /**
   * The time.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Constructs a new PaymentGatewayBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \GuzzleHttp\ClientInterface $client
   *   The client.
   * @param \Drupal\commerce_price\RounderInterface $rounder
   *   The price rounder.
   */


  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, PaymentTypeManager $payment_type_manager, PaymentMethodTypeManager $payment_method_type_manager, TimeInterface $time, ClientInterface $client, RounderInterface $rounder, MessengerInterface $messenger) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);

    $this->httpClient = $client;
    $this->rounder = $rounder;
    $this->messenger = $messenger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('http_client'),
      $container->get('commerce_price.rounder'),
      $container->get('messenger')
    );
  }


  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
        'public_key' => '',
        'private_key' => '',
        'shipping_info' => FALSE,
        'merchant_completes' => FALSE,
        'api_logging' => FALSE,
        'skip_review' => FALSE,
      ] + parent::defaultConfiguration();
  }

  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['public_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Public key'),
      '#description' => $this->t('This is the public key from the SezzlePay checkout.'),
      '#default_value' => $this->configuration['public_key'],
      '#required' => TRUE,
    ];

    $form['private_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Private key'),
      '#description' => $this->t('This is the private key from the SezzlePay checkout.'),
      '#default_value' => $this->configuration['private_key'],
      '#required' => TRUE,
    ];

    $form['shipping_info'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Collect shipping info on Sezzle Pay'),
      '#description' => $this->t('Optional flag to indicate if you would like Sezzle Pay to collect shipping information for  customer, default is FALSE'),
      '#default_value' => $this->configuration['shipping_info'],
      '#required' => FALSE,
    ];

    $form['merchant_completes'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Authorize only on Sezzle pay, complete checkout on Drupal site'),
      '#description' => $this->t('Optional flag to determine whether checkout will capture funds or only authorize, default is FALSE and that will capture funds, with TRUE, you will have to add extra code to complete order and capture funds, more info about this can be found in readme'),
      '#default_value' => $this->configuration['merchant_completes'],
      '#required' => FALSE,
    ];

    $form['skip_review'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Option to skip review step(pane) and go straight to Sezzle site, applicable only when sezzle payment selected'),
      '#default_value' => $this->configuration['skip_review'],
      '#required' => FALSE,
    ];

    $form['api_logging'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Log webhook messages onNotify for debugging'),
      '#default_value' => $this->configuration['api_logging'],
      '#required' => FALSE,
    ];


    return $form;
  }

  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    // Format webhook url and send it as config on submit
    $webhook_url = Url::fromRoute('commerce_payment.notify', ['commerce_payment_gateway' => $form_state->getValue("id"),], ['absolute' => TRUE])
      ->toString();

    $values = $form_state->getValue($form['#parents']);
    $this->configuration['public_key'] = $values['public_key'];
    $this->configuration['private_key'] = $values['private_key'];
    $this->configuration['webhook_url'] = $webhook_url;
    $this->configuration['shipping_info'] = $values['shipping_info'];
    $this->configuration['merchant_completes'] = $values['merchant_completes'];
    $this->configuration['api_logging'] = $values['api_logging'];
    $this->configuration['skip_review'] = $values['skip_review'];

    // Send webhook config to Sezzle and save it as form is saved
    $keys['public_key'] = $values['public_key'];
    $keys['private_key'] = $values['private_key'];
    $sezzle_auth_response = $this->doAuthentication($keys);
    $this->setWebHookConfig($webhook_url, $sezzle_auth_response['token']);
  }

  /**
   * {@inheritdoc}
   */
  public function setWebHookConfig($webhook_url, $token) {
    if ($this->getMode() == 'test') {
      $url = 'https://sandbox.gateway.sezzle.com/v1/configuration';
    }
    else {
      $url = 'https://gateway.sezzle.com/v1/configuration';
    }

    $request = $this->httpClient->POST($url, [
      'headers' => [
        'Authorization' => $token,
      ],
      'json' => [
        'webhook_url' => $webhook_url,
      ],
    ]);
    // There is no response body
    $response = json_decode($request->getStatusCode());

    return $response;
  }

  /**
   * {@inheritdoc}
   */
  public function getRedirectUrl($order_id) {
    if ($this->getMode() == 'test') {
      return 'https://sandbox.checkout.sezzle.com/?id=' . $order_id;
    }
    else {
      return 'https://checkout.sezzle.com/?id=' . $order_id;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getOrderDetails($order_id, $token, $include_shipping = TRUE) {
    if ($this->getMode() == 'test') {
      $url = 'https://sandbox.gateway.sezzle.com/v1/orders/' . $order_id;
    }
    else {
      $url = 'https://gateway.sezzle.com/v1/orders/' . $order_id;
    }

    $request = $this->httpClient->get($url, [
      'headers' => [
        'Authorization' => $token,
      ],
      'json' => [
        'include-shipping-info' => $include_shipping,
      ],
    ]);
    $response = json_decode($request->getBody(), TRUE);

    return $response;
  }

  /**
   * {@inheritdoc}
   */
  public function getCheckoutUrl() {
    if ($this->getMode() == 'test') {
      return 'https://sandbox.gateway.sezzle.com/v1/checkouts';
    }
    else {
      return 'https://gateway.sezzle.com/v1/checkouts';
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getApiAuthenticationUrl() {
    if ($this->getMode() == 'test') {
      return 'https://sandbox.gateway.sezzle.com/v1/authentication';
    }
    else {
      return 'https://gateway.sezzle.com/v1/authentication';
    }
  }

  /**
   * {@inheritdoc}
   */
  public function doAuthentication($keys) {
    // Get SezzlePay Token TODO maybe use try/catch here
    $request = $this->httpClient->POST($this->getApiAuthenticationUrl(), [
      'json' => $keys,
    ]);
    $response = json_decode($request->getBody(), TRUE);

    return $response;
  }

  /**
   * {@inheritdoc}
   */
  public function completeCheckout($order_id, $token) {
    if ($this->getMode() == 'test') {
      $url = 'https://sandbox.gateway.sezzle.com/v1/checkouts/' . $order_id . '/complete';
    }
    else {
      $url = 'https://gateway.sezzle.com/v1/checkouts/' . $order_id . '/complete';
    }

    $request = $this->httpClient->post($url, [
      'headers' => [
        'Authorization' => $token,
      ],
    ]);
    $response = json_decode($request->getBody(), TRUE);

    return $response;
  }

  /**
   * {@inheritdoc}
   */
  public function mapAndCreateSezzleOrder(PaymentInterface $payment, array $extra, $token) {
    $order = $payment->getOrder();
    $amount = $this->rounder->round($payment->getAmount());
    $configuration = $this->getConfiguration();


    if ($extra['merchant_completes'] == 1) {
      $merchant_completes = TRUE;
    }
    else {
      $merchant_completes = FALSE;
    }

    if ($extra['shipping_info'] == 1) {
      $requires_shipping_info = TRUE;
    }
    else {
      $requires_shipping_info = FALSE;
    }


    // Prepare basic data
    $body_data = [
      'amount_in_cents' => $this->toMinorUnits($amount),
      'currency_code' => $amount->getCurrencyCode(),
      'order_description' => 'Replace below with item names',
      'order_reference_id' => $order->order_id->getString(),
      'checkout_cancel_url' => $extra['cancel_url'],
      'checkout_complete_url' => $extra['return_url'],
    ];


    // Prepare customer details from user profile or billing data
    if ($profile = $order->getCustomer()) {

      // Check if this is anonymous user
      if ($profile->uid->getString() != 0) {
        $customer_details = [
          'first_name' => isset($profile->field_first_name) ? $profile->field_first_name->getString() : '',
          'last_name' => isset($profile->field_last_name) ? $profile->field_last_name->getString() : '',
          'email' => isset($profile->mail) ? $profile->mail->getString() : '',
          'phone' => isset($profile->field_phone) ? $profile->field_phone->getString() : '',
        ];
        // Filter out empty values.
        $body_data += ["customer_details" => array_filter($customer_details)];
      }
    }

    // Prepare shipping address, check if the order references shipments.
    if ($order->hasField('shipments') && !$order->get('shipments')->isEmpty()) {
      // Gather the shipping profiles and send shipping information
      $shipping_profiles = [];

      // Loop over the shipments to collect shipping profiles.
      foreach ($order->get('shipments')->referencedEntities() as $shipment) {
        if ($shipment->get('shipping_profile')->isEmpty()) {
          continue;
        }
        $shipping_profile = $shipment->getShippingProfile();
        $shipping_profiles[$shipping_profile->id()] = $shipping_profile;
      }

      // If there are more shipping profiles we will use first one
      $shipping_profile = reset($shipping_profiles);
      /** @var \Drupal\address\AddressInterface $address */
      $address = $shipping_profile->address->first();
      $name = $address->getGivenName() . ' ' . $address->getFamilyName();
      $shipping_address = [
        'name' => substr($name, 0, 32),
        'street' => substr($address->getAddressLine1(), 0, 100),
        'street2' => substr($address->getAddressLine2(), 0, 100),
        'city' => substr($address->getLocality(), 0, 40),
        'state' => substr($address->getAdministrativeArea(), 0, 40),
        'country_code' => $address->getCountryCode(),
        'postal_code' => substr($address->getPostalCode(), 0, 20),
        'phone_number' => $shipping_profile->field_phone->getString(),
      ];
      // Filter out empty values.
      $body_data += ["shipping_address" => array_filter($shipping_address)];
    }

    // Prepare billing address
    if ($order->getBillingProfile()) {

      $address = $order->getBillingProfile()->address->first();
      $first_name = $address->getGivenName();
      $last_name = $address->getFamilyName();
      $name = $first_name . ' ' . $last_name;
      $order_email = $order->getEmail();
      $phone = isset($order->getBillingProfile()->field_phone) ? $order->getBillingProfile()->field_phone->getString() : '';

      $billing_address = [
        'name' => substr($name, 0, 32),
        'street' => substr($address->getAddressLine1(), 0, 100),
        'street2' => substr($address->getAddressLine2(), 0, 100),
        'city' => substr($address->getLocality(), 0, 40),
        'state' => substr($address->getAdministrativeArea(), 0, 40),
        'country_code' => $address->getCountryCode(),
        'postal_code' => substr($address->getPostalCode(), 0, 20),
        'phone_number' => $phone,
      ];
      // Filter out empty values.
      $body_data += ["billing_address" => array_filter($billing_address)];

      // If this was anonymous user, fill in billing data
      if ($profile->uid->getString() == 0) {
        $customer_details = [
          'first_name' => isset($first_name) ? $first_name : '',
          'last_name' => isset($last_name) ? $last_name : '',
          'email' => isset($order_email) ? $order_email : '',
          'phone' => isset($phone) ? $phone : '',
        ];
        // Filter out empty values.
        $body_data += ["customer_details" => array_filter($customer_details)];
      }
    }


    // Flag that makes Sezzle pay collect shipping information from the customer
    $body_data += ["requires_shipping_info" => $requires_shipping_info];

    $currency_code = $amount->getCurrencyCode();
    // Calculate the items total.
    $items_total = new Price('0', $currency_code);

    $all_items = [];
    $order_description = "";
    // Add order item data.
    foreach ($order->getItems() as $item) {
      $product = ProductVariation::load($item->getPurchasedEntityId());
      $all_items[] = [
        'name' => $item->getTitle(),
        'sku' => $product->getSku(),
        'quantity' => round($item->getQuantity(), 0),
        "price" => [
          'amount_in_cents' => $this->toMinorUnits($item->getTotalPrice()),
          'currency' => $item->getTotalPrice()->getCurrencyCode(),
        ],
      ];
      $order_description .= $item->getTitle() . ' ' . 'x ' . round($item->getQuantity(), 0) . ', ';
    }
    $body_data += ["items" => array_filter($all_items)];
    $body_data['order_description'] = $this->t('Order has following items: ') . substr($order_description, 0, -2);


    // Initialize Shipping,Tax and Discounte prices, they need to be sent separately to Sezzle pay.
    $shipping_amount_init = new Price('0', $currency_code);
    $tax_amount_init = new Price('0', $currency_code);

    // Collect the adjustments (Shipping, Taxes and Discounts)
    $adjustments = [];
    foreach ($order->collectAdjustments() as $adjustment) {
      // Skip included adjustments.
      if ($adjustment->isIncluded()) {
        continue;
      }
      // Tax & Shipping adjustments need to be handled separately.
      if ($adjustment->getType() == 'shipping') {
        $shipping_amount_init = $shipping_amount_init->add($adjustment->getAmount());
      }
      // Add taxes that are not included in the items total.
      elseif ($adjustment->getType() == 'tax') {
        $tax_amount_init = $tax_amount_init->add($adjustment->getAmount());
      }
      else {
        // Collect other adjustments.
        $type = $adjustment->getType();
        $source_id = $adjustment->getSourceId();
        if (empty($source_id)) {
          // Adjustments without a source ID are always shown standalone.
          $key = count($adjustments);
        }
        else {
          // Adjustments with the same type and source ID are combined.
          $key = $type . '_' . $source_id;
        }

        if (empty($adjustments[$key])) {
          $adjustments[$key] = [
            'type' => $type,
            'label' => (string) $adjustment->getLabel(),
            'total' => $adjustment->getAmount(),
          ];
        }
        else {
          $adjustments[$key]['total'] = $adjustments[$key]['total']->add($adjustment->getAmount());
        }
      }
    }

    if (!empty($adjustments)) {
      $discounts = [];
      // Setup adjustments where disocounts
      foreach ($adjustments as $adjustment) {
        if ($adjustment['type'] == 'promotion') {
          $adjustment_amount = $this->rounder->round($adjustment['total']);
          $discounts[] = [
            'name' => $adjustment['label'],
            'amount' => [
              'amount_in_cents' => $this->toMinorUnits($adjustment_amount),
              'currency' => $adjustment_amount->getCurrencyCode(),
            ],
          ];
        }
      }
      $body_data += ["discounts" => array_filter($discounts)];
    }


    // Send the shipping amount separately.
    if (!$shipping_amount_init->isZero()) {
      $shipping_amount_init = $this->rounder->round($shipping_amount_init);

      $shipping_amount['amount_in_cents'] = $this->toMinorUnits($shipping_amount_init);
      $shipping_amount['currency'] = $shipping_amount_init->getCurrencyCode();

      $body_data += ["shipping_amount" => $shipping_amount];
    }


    // Send the tax amount.
    if (!$tax_amount_init->isZero()) {
      $tax_amount_init = $this->rounder->round($tax_amount_init);
      $tax_amount['amount_in_cents'] = $this->toMinorUnits($tax_amount_init);
      $tax_amount['currency'] = $tax_amount_init->getCurrencyCode();

      $body_data += ["tax_amount" => $tax_amount];
    }

    // Flag that makes customer return to Drupal site and finish order there
    $body_data += ["merchant_completes" => $merchant_completes];

    // Create order in Sezzle system, get the order URL to redirect user to it
    try {
      $request = $this->httpClient->post($this->getCheckoutUrl(), [
        'json' => $body_data,
        'headers' => [
          'Authorization' => $token,
        ],
      ]);
      $response = json_decode($request->getBody());
    } catch (RequestException $e) {
      watchdog_exception('commerce_sezzle_pay', $e->getMessage());
      //TODO make some redirects if exception so user lands on normal page and not error page
    }

    return $response;
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {

    $payment_gateway_plugin = $order->get('payment_gateway')->first()->entity;

    /* Sezzle doesnt return anything when sending user back to drupal site, so we
     * need to make an API call and get data from it. We only do this if
     * merchant_completes is set to FALSE, if TRUE, then order needs to be set to
     * Complete somewhere else, on some manual or auto approve action, we skip
     * the rest if merchant_completes == 1
     */

    if ($payment_gateway_plugin->get("configuration")['merchant_completes']) {
      return;
    }

    $data = [];
    $data['public_key'] = $payment_gateway_plugin->get("configuration")['public_key'];
    $data['private_key'] = $payment_gateway_plugin->get("configuration")['private_key'];

    $sezzle_auth_response = $this->doAuthentication($data);
    $token = $sezzle_auth_response['token'];
    $order_id = $order->order_id->getString();

    // Run complete order action, to capture the funds, this needs to be run somewhere else
   // $complete_checkout = $this->completeCheckout($order_id, $token);

    // Get sezzle order data, it will be null if it is not captured
    $sezzle_order = $this->getOrderDetails($order_id, $token);

    if ($sezzle_order["captured_at"] != NULL) {
      $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
      $payment = $payment_storage->create([
        'state' => 'completed',
        'amount' => $order->getTotalPrice(),
        'payment_gateway' => $this->entityId,
        'order_id' => $order->id(),
      ]);

      $payment->save();
    }
    else {
      throw new PaymentGatewayException("Sezzle pay had an error", $sezzle_order);
    }

  }

  public function onNotify(Request $request) {

    // Log the response message if request logging is enabled.
    if (!empty($this->configuration['api_logging'])) {
      \Drupal::logger('commerce_sezzle_pay')
        ->debug('e-Commerce notification: <pre>@body</pre>', [
          '@body' => var_export($request->query->all(), TRUE),
        ]);
    }

    // Get Webhook request data.
    $webhook_data = $this->getRequestDataArray($request->getContent());
    //TODO get this data somehow and then act on it, make refund or order update
    // https://docs.drupalcommerce.org/commerce2/developer-guide/payments/create-payment-gateway/on-site-gateways/creating-payments

  }

  /**
   * Get data array from a request content.
   *
   * @param string $request_content
   *   The Request content.
   *
   * @return array
   *   The request data array.
   */
  protected function getRequestDataArray($request_content) {
    parse_str(html_entity_decode($request_content), $webhook_data);
    return $webhook_data;
  }

  /**
   * {@inheritdoc}
   */
  public function onCancel(OrderInterface $order, Request $request) {
    $this->messenger->addMessage($this->t('You have canceled checkout at @gateway but may resume the checkout process here when you are ready.', [
      '@gateway' => $this->getDisplayLabel(),
    ]));
  }
}